<script type="text/javascript">
    $(function(){
        $('#appsearch-form').appsearch({
            'formChoices': {{ search.render_all_constraint_choices }},
                
            // Passed directly to formset.js
            'formsetOptions': {
                prefix: '{{ search.constraint_formset.prefix }}',
                addText: '',
                deleteText: ''
            }
        });
    });
</script>

{{ search.render_search_form }}
    
<div>&nbsp;</div>

{{ search.render_results_list }}
